from fnmatch import fnmatch
from enum import Enum
from typing import Optional, Container, FrozenSet, Iterable, List, Dict, Union

from pydantic import BaseModel

__all__ = [
    "Feature",
    "FeatureProperties",
    "FeatureTask",
    "ManifestInfo",
    "Modpack",
    "ManifestIndex",
    "ModpackStatus",
    "PatternTestable",
    "RichManifest",
    "SimplePatternList",
    "SourceManifest",
    "TargetManifest",
    "Task",
    "UserFileTask",
]


class BaseManifest(BaseModel):
    name: str
    title: str
    version: str = "1.0"

    class Config:
        allow_mutation = False

    def __hash__(self):
        return hash((self.name, self.title, self.version))

    def __eq__(self, other):
        return hash(self) == hash(other)


class ManifestInfo(BaseManifest):
    location: str

    class Config:
        allow_mutation = False

    def __hash__(self):
        return hash((self.name, self.title, self.version, self.location))

    def __eq__(self, other):
        return hash(self) == hash(other)


class ManifestIndex(BaseModel):
    minimumVersion: int = 3
    packages: List[ManifestInfo] = []

    class Config:
        allow_mutation = False

    def __hash__(self):
        return hash((self.minimumVersion, frozenset(self.packages)))

    def __eq__(self, other):
        return hash(self) == hash(other)


class SimplePatternList(BaseModel):
    include: List[str] = []
    exclude: List[str] = []

    def __hash__(self):
        return hash(frozenset(self.include + self.exclude))

    def __eq__(self, other):
        return hash(self) == hash(other)

    class Config:
        allow_mutation = False


class FeatureProperties(BaseModel):
    name: str
    description: str
    recommendation: str = "starred"
    selected: bool = False

    def __hash__(self):
        return hash((self.name, self.description, self.recommendation, self.selected))

    def __eq__(self, other):
        return hash(self) == hash(other)

    class Config:
        allow_mutation = False


class Feature(BaseModel):
    properties: FeatureProperties
    files: SimplePatternList

    class Config:
        allow_mutation = False


class Task(BaseModel):
    type: str = "file"
    hash: str
    location: str
    to: str
    size: int

    def __hash__(self):
        return hash((
            self.type,
            self.hash,
            self.location,
            self.to,
            self.size,
        ))

    def __eq__(self, other):
        return hash(self) == hash(other)

    class Config:
        allow_mutation = False
        ignore_extra = False


class UserFileTask(BaseModel):
    type: str = "file"
    hash: str
    location: str
    to: str
    size: int

    userFile: bool = True

    def __hash__(self):
        return hash((
            self.type,
            self.hash,
            self.location,
            self.to,
            self.size,
            self.userFile,
        ))

    def __eq__(self, other):
        return hash(self) == hash(other)

    class Config:
        allow_mutation = False
        ignore_extra = False


class FeatureTask(BaseModel):
    type: str = "file"
    hash: str
    location: str
    to: str
    size: int

    when: Dict

    def __hash__(self):
        return hash((
            self.type,
            self.hash,
            self.location,
            self.to,
            self.size,
            self.when['if'],
            frozenset(self.when['features']),
        ))

    def __eq__(self, other):
        return hash(self) == hash(other)

    class Config:
        allow_mutation = False
        ignore_extra = False


class RichManifest(BaseManifest):
    gameVersion: str
    launch: Dict[str, List[str]] = {"flags": []}
    minimumVersion: int = 2
    librariesLocation: Optional[str] = None
    objectsLocation: str
    userFiles: SimplePatternList = SimplePatternList()

    class Config:
        allow_mutation = False


class SourceManifest(RichManifest):
    features: List[Feature] = []

    class Config:
        allow_mutation = False
        ignore_extra = False


class TargetManifest(RichManifest):
    features: List[FeatureProperties] = []
    tasks: List[Union[Task, UserFileTask, FeatureTask]] = []

    def __hash__(self):
        return hash((
            self.gameVersion,
            frozenset(self.launch['flags']),
            self.userFiles,
            self.minimumVersion,
            self.librariesLocation,
            self.objectsLocation,
            frozenset(self.features),
            frozenset(self.tasks),
        ))

    def __eq__(self, other):
        return hash(self) == hash(other)

    class Config:
        allow_mutation = False
        ignore_extra = False


class ModpackStatus(Enum):
    FRESH = "Fresh"
    OUTDATED = "Outdated"
    NOT_DEPLOYED = "Not deployed"
    ORPHANED = "Orphaned"
    BROKEN = "Broken"
    UNKNOWN = "Unknown"


class Modpack(BaseModel):
    name: str
    enabled: bool
    status: ModpackStatus
    source_manifest: Optional[SourceManifest] = None
    target_manifest: Optional[TargetManifest] = None

    class Config:
        allow_mutation = False
        ignore_extra = False


class PatternTestable(Container):
    __patterns: FrozenSet
    __slots__ = ('__patterns',)

    def __contains__(self, item: str) -> bool:
        return any(fnmatch(item, pattern) for pattern in self.__patterns)

    def __init__(self, patterns: Iterable[str]):
        self.__patterns = frozenset(patterns)

    def __len__(self) -> int:
        return len(self.__patterns)



from typing import List
from pathlib import Path

import click

from apeiron.bootstrap import check
from apeiron.constants import OBJECTS_DIR
from apeiron.exceptions import ManifestError
from apeiron.exceptions import ModpackError
from apeiron.manifest import bootstrap_source_manifest
from apeiron.models import SourceManifest
from apeiron.modpack import get_modpack, get_modpacks, enable_modpack, build_modpack, disable_modpack, delete_modpack
from apeiron.view import print_modpack, print_modpacks, print_source_manifest


@click.group()
@click.option(
    '--config',
    expose_value=True,
    default=None,
    help="Additional cfg",
    type=click.Path(exists=True, resolve_path=True)
)
def cli(config: str):
    """Simple CLI for modpack management"""
    check(config)


@cli.command(name="list")
def _list():
    """Lists all modpacks"""

    click.echo(click.style("Available modpacks", fg="blue", bold=True))
    click.echo(print_modpacks(get_modpacks()))


@cli.command(name="show")
@click.argument("name", metavar="<name>")
def show(name: str):
    """Show detailed info about modpack <name>"""

    click.echo(print_modpack(get_modpack(name)))


@cli.command(name="create")
def create():
    """Create new empty modpack"""

    name: str = click.prompt("Enter modpack name (internal identifier)")
    title: str = click.prompt("Enter modpack displayed title", default=name.capitalize())
    game_version: str = click.prompt("Enter modpack version", default="1.10.2")
    libraries_url: str = click.prompt("Specify libraries URL", default=None)
    objects_dir: str = click.prompt("Specify objects location for this modpack", default=OBJECTS_DIR)

    source_manifest: SourceManifest = SourceManifest(
        name=name,
        title=title,
        gameVersion=game_version,
        librariesLocation=libraries_url,
        objectsLocation=objects_dir
    )

    click.echo()
    click.echo(click.style("Check generated modpack stub below:", bold=True))
    click.echo(print_source_manifest(source_manifest))

    if not click.confirm("Is this correct?"):
        return click.echo(click.style("Aborted", fg='red'))

    try:
        path: Path = bootstrap_source_manifest(source_manifest)
    except ManifestError as e:
        click.echo(click.style(str(e), fg="red"))
    else:
        click.echo(click.style("Bootstrapped new modpack!", fg="green", bold=True))
        click.echo(f"You can now edit modpack details in {click.style(path.as_posix(), bold=True)}")


@cli.command(name="enable")
@click.argument("name", metavar="<name>")
def enable(name: str):
    """Enables modpack <name>"""

    click.echo(f"Enabling modpack «{click.style(name, bold=True)}»")
    try:
        enable_modpack(name)
    except ModpackError as e:
        click.echo(click.style(str(e), fg="yellow"))
    else:
        click.echo(click.style(f"Enabled modpack «{click.style(name, bold=True)}»", fg="green"))


@cli.command(name="disable")
@click.argument("name", metavar="<name>")
def disable(name: str):
    """Disables modpack <name>"""

    click.echo(f"Disabling modpack «{click.style(name, bold=True)}»")
    try:
        disable_modpack(name)
    except ModpackError as e:
        click.echo(click.style(str(e), fg="yellow"))
    else:
        click.echo(click.style(f"Disabled modpack «{click.style(name, bold=True)}»", fg="green"))


@cli.command(name="delete")
@click.option('--prune', is_flag=True, expose_value=True, default=False, help="Prunes sources as well as built files")
@click.argument("name", metavar="<name>")
@click.pass_context
def delete(ctx: click.core.Context, name: str, prune: bool):
    """Deletes modpack <name>"""

    ctx.invoke(disable, name=name)
    click.echo(f"Deleting files for modpack «{click.style(name, bold=True)}»")
    try:
        result: List[ManifestError] = delete_modpack(name, prune)
    except ModpackError as e:
        click.echo(click.style(str(e), fg='red'))
    else:
        for exc in result:
            click.echo(click.style(str(exc), fg='yellow'))
        if prune:
            click.echo(click.style(f"Completely removed modpack «{click.style(name, bold=True)}»", fg="green"))
        else:
            click.echo(click.style(f"Deleted built files for modpack «{click.style(name, bold=True)}»", fg="green"))


@cli.command(name="build")
@click.option('--debug', is_flag=True, expose_value=True, default=False, help="Print some debug info")
@click.option('--strict', is_flag=True, expose_value=True, default=False, help="Strict check for pre-existing files")
@click.argument("name", metavar="<name>")
@click.pass_context
def build(ctx: click.core.Context, name: str, strict: bool, debug: bool):
    """Build modpack <name>"""

    click.echo(f"Building modpack «{click.style(name, bold=True)}»…")
    try:
        build_modpack(name, strict, debug)
    except ModpackError as e:
        click.echo(click.style(str(e), fg="yellow"))
    except ManifestError as e:
        click.echo(click.style(str(e), fg="red"))
    else:
        click.echo(click.style(f"Successfully build modpack «{click.style(name, bold=True)}»", fg="green"))
        ctx.invoke(show, name=name)


if __name__ == '__main__':
    cli()



from pathlib import Path

from yamlcfg import YAMLConfig


cfg = YAMLConfig(
    paths=[
        Path('~/.config/apeiron.yaml').expanduser().as_posix(),
        Path('/etc/config/apeiron.yaml').as_posix(),
    ],
    permute=False
)

DEFAULTS = dict(
    storage_dir=Path('~/apeiron/storage').expanduser().as_posix(),
    modpack_index="index.json",
    parallelism=12,
)



from typing import NamedTuple, List, Tuple

from tabulate import tabulate
from click import style
from yamlcfg import YamlConfig

from apeiron.models import Modpack
from apeiron.models import ModpackStatus
from apeiron.models import SourceManifest


class ShortModpackInfo(NamedTuple):
    id: str
    title: str
    status: str
    enabled: str


STATUS_COLOR_MAP = {
    ModpackStatus.FRESH: dict(fg="green", bold=True),
    ModpackStatus.OUTDATED: dict(fg="yellow"),
    ModpackStatus.NOT_DEPLOYED: dict(fg="green", dim=True),
    ModpackStatus.BROKEN: dict(fg="red", bold=True),
    ModpackStatus.UNKNOWN: dict(fg="black"),
    ModpackStatus.ORPHANED: dict(fg="red", dim=True),
}

ACTIVITY_COLOR_MAP = {
    True: dict(bold=True),
    False: dict(dim=True),
}


def extract_info(modpack: Modpack) -> ShortModpackInfo:
    if modpack.source_manifest is not None:
        title = modpack.source_manifest.title
    elif modpack.target_manifest is not None:
        title = modpack.target_manifest.title
    else:
        title = "???"

    return ShortModpackInfo(
        id=style(modpack.name, bold=modpack.enabled),
        title=title,
        status=style(modpack.status.value, **STATUS_COLOR_MAP[modpack.status]),
        enabled=style("Yes" if modpack.enabled else "No", **ACTIVITY_COLOR_MAP[modpack.enabled]),
    )


def print_modpacks(modpacks: List[Modpack]) -> str:
    return tabulate(
        map(extract_info, modpacks),
        headers=[style(s.capitalize(), fg='blue', bold=True) for s in ShortModpackInfo._fields],
        tablefmt="fancy_grid"
    )


def print_source_manifest(manifest: SourceManifest) -> str:
    user_files: int = len(manifest.userFiles.include)
    features: int = len(manifest.features)

    data: List[Tuple[str, str]] = [
        ("Name", style(manifest.name, fg='green')),
        ("Displayed title", style(manifest.title, fg='green')),
        ("Game version", style(manifest.gameVersion, fg='green')),
        ("Objects location", style(manifest.objectsLocation, fg='green')),
        ("User-managed files", style(f"{user_files} file(s)") if user_files else style("n/a", dim=True)),
        ("Enabled features", style(f"{features} feature(s)") if features else style("n/a", dim=True)),
    ]
    if manifest.librariesLocation is not None:
        data.append(("Libraries location", style(manifest.librariesLocation, fg='green')))

    return tabulate(data, tablefmt="fancy_grid")


def print_config(cfg: YamlConfig) -> str:
    # noinspection PyProtectedMember
    return tabulate(
        cfg._data.items(),
        headers=[style("Key", bold=True), style("Value", bold=True)],
        tablefmt="fancy_grid"
    )


def print_modpack(modpack: Modpack) -> str:
    short_info: ShortModpackInfo = extract_info(modpack)
    header = style(f"Package {short_info.id} ({short_info.title})", fg="yellow", bold=True)
    info = f"Status: {short_info.status} Enabled: {short_info.enabled}"
    source_manifest_info: str = print_source_manifest(modpack.source_manifest)
    return "\n".join((header, info, source_manifest_info))



import sys
from pathlib import Path
from setuptools import setup, find_packages

if not sys.version_info >= (3, 6):
    raise RuntimeError('apeiron requires at least python 3.6')

readme = Path('README.rst').read_text()
history = Path('HISTORY.rst').read_text()

requirements = Path('requirements.txt').read_text().splitlines()
setup_requirements = []
test_requirements = Path('requirements_dev.txt').read_text().splitlines()

setup(
    author='Andriy Kushnir (Orhideous)',
    author_email='me@orhideous.name',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description='Simple CLI for modpack management',
    install_requires=requirements,
    license='MIT license',
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords=['Minecraft', 'modpack', 'skcraft'],
    name='apeiron',
    packages=find_packages(include=['apeiron']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/Orhideous/apeiron',
    version='0.9.1',
    zip_safe=False,
    entry_points='''
        [console_scripts]
        apeiron=apeiron.cli:cli
    ''',
)



class ModpackError(Exception):
    pass


class ManifestError(Exception):
    pass


class InvalidStatusError(ModpackError):
    pass


class StateConflictError(ModpackError):
    pass


class ConsistenceError(ManifestError):
    pass



from pathlib import Path
from typing import Optional

import click
from yaml import YAMLError
from yamlcfg import YAMLConfig

from apeiron.config import cfg
from apeiron.config import DEFAULTS
from apeiron.constants import SOURCES_DIR, PACKAGES_DIR
from apeiron.models import ManifestIndex
from apeiron.manifest import get_manifest_index_path
from apeiron.manifest import save_manifests


from apeiron.view import print_config


def create_dir_if_not_exist(path: Path, name: str):
    if not path.exists():
        click.echo(click.style(f"{name.capitalize()} not found, creating…", fg="yellow"))
        try:
            path.mkdir(parents=True)
        except IOError as e:
            click.echo(click.style(f"Failed to create {name}, reason: {str(e)}", fg="red"))
            exit(1)
        else:
            click.echo(click.style(f"Created empty {name}: {path.as_posix()}", fg="green"))


# noinspection PyProtectedMember
def check(extra_config_path: Optional[str]):
    local_config: Path = Path(cfg._paths[0])
    if not local_config.exists():
        click.echo(click.style(f"Config not found, writing defaults…", fg="yellow"))
        cfg._data = DEFAULTS
        try:
            cfg.write()
        except IOError as e:
            click.echo(click.style(f"Failed to write default config, reason: {str(e)}", fg="red"))
            exit(1)
        else:
            click.echo(click.style(f"Wrote defaults to: {local_config.as_posix()}", fg="green"))
    if extra_config_path is not None:
        try:
            augmented_config: YAMLConfig = YAMLConfig(path=extra_config_path)
        except YAMLError as e:
            click.echo(click.style(
                f"Failed to update config with values from {extra_config_path}, reason: {str(e)}", fg="red"
            ))
            exit(1)
        else:
            cfg._data.update(augmented_config._data)
            click.echo(click.style(f"Loaded extra config from: {extra_config_path}", fg="green"))
    click.echo(click.style("Application config", fg="blue", bold=True))
    click.echo(print_config(cfg))

    base_storage: Path = Path(cfg.storage_dir)
    create_dir_if_not_exist(base_storage, "base storage")

    source_dir: Path = base_storage / SOURCES_DIR
    create_dir_if_not_exist(source_dir, "source dir")

    package_dir: Path = base_storage / PACKAGES_DIR
    create_dir_if_not_exist(package_dir, "package dir")

    manifest_index_path: Path = get_manifest_index_path()
    if not manifest_index_path.exists():
        click.echo(click.style(f"Manifests index not found, writing stub…", fg="yellow"))
        try:
            save_manifests(ManifestIndex())
        except IOError as e:
            click.echo(click.style(f"Failed to write manifests index, reason: {e}", fg="red"))
            exit(1)
        else:
            click.echo(click.style(f"Wrote empty manifest index to: {manifest_index_path.as_posix()}", fg="green"))



import json
from typing import Optional, Dict, Union, List, Any, Iterable

import yaml
import uuid
from shutil import rmtree
from shutil import copy
from functools import partial
from multiprocessing import cpu_count
from multiprocessing import Pool
from hashlib import sha1
from pathlib import Path

import click
from pydantic import ValidationError

from apeiron import constants as c
from apeiron import models as m
from apeiron.config import cfg
from apeiron.exceptions import ManifestError


def validate_manifest_fs(package_id: str, manifest: m.RichManifest) -> bool:
    consistent = manifest.name == package_id
    if not consistent:
        click.echo(click.style(
            f"Inconsistent {manifest.__class__.__name__}! "
            f"Internal id equals «{manifest.name}», but resides in «{package_id}»!",
            fg='red',
            bold=True
        ))
    return consistent


def load_target_manifest(package_id: str) -> Optional[m.TargetManifest]:
    path: Path = Path(cfg.storage_dir) / c.PACKAGES_DIR / package_id / c.TARGET_MANIFEST_INDEX

    if not path.exists():
        return None

    try:
        manifest = m.TargetManifest(**json.load(path.open()))
    except ValidationError as e:
        click.echo(click.style(f"Broken TargetManifest, id: «{package_id}». Reason: {str(e)}", fg='red'))
        return None
    else:
        return validate_manifest_fs(package_id, manifest) and manifest or None


def _validate_task(path: Path, task: m.Task) -> bool:
    return (path / task.location).exists()


def validate_target_manifest(manifest: m.TargetManifest) -> bool:
    path: Path = Path(cfg.storage_dir) / c.PACKAGES_DIR / manifest.name / manifest.objectsLocation

    if not path.exists():
        return False

    with Pool(processes=cfg.parallelism or cpu_count()) as pool:
        return all(pool.imap_unordered(partial(_validate_task, path), manifest.tasks))


def delete_target_manifest(manifest: m.TargetManifest):
    path: Path = Path(cfg.storage_dir) / c.PACKAGES_DIR / manifest.name
    rmtree(path.as_posix())


def load_source_manifest(package_id: str) -> Optional[m.SourceManifest]:
    path: Path = Path(cfg.storage_dir) / c.SOURCES_DIR / package_id / c.SOURCE_MANIFEST_INDEX

    if not path.exists():
        return None

    try:
        manifest = m.SourceManifest(**yaml.load(path.open()))
    except ValidationError as e:
        click.echo(click.style(f"Broken SourceManifest, id: «{package_id}». Reason: {str(e)}", fg='red'))
        return None
    else:
        return validate_manifest_fs(package_id, manifest) and manifest or None


def delete_source_manifest(manifest: m.SourceManifest):
    path: Path = Path(cfg.storage_dir) / c.SOURCES_DIR / manifest.name
    rmtree(path.as_posix())


def _generate_task(
    user_files: m.PatternTestable,
    features: Dict[str, m.PatternTestable],
    base_dir: Path,
    path: Path,
) -> Union[m.Task, m.UserFileTask, m.FeatureTask]:
    with path.open(mode='rb') as file:
        data: str = file.read()
        digest: str = sha1(data).hexdigest()
        destination: str = path.relative_to(base_dir).as_posix()
        affected_features: List[str] = [
            name
            for name, feature_pattern
            in features.items()
            if destination in feature_pattern
        ]
        base_data: Dict[str, Union[str, int]] = dict(
            hash=digest,
            location="{}/{}/{}".format(digest[0:2], digest[2:4], digest),
            to=destination,
            size=len(data)
         )
        if affected_features:
            return m.FeatureTask(
                when={"if": "requireAny", "features": affected_features},
                **base_data
            )
        elif destination in user_files:
            return m.UserFileTask(**base_data)
        else:
            return m.Task(**base_data)


def build_target_manifest(source: m.SourceManifest) -> m.TargetManifest:
    user_files: m.PatternTestable = m.PatternTestable(source.userFiles.include)
    features: Dict[str, m.PatternTestable] = {
        f.properties.name: m.PatternTestable(f.files.include)
        for f
        in source.features
    }

    path: Path = Path(cfg.storage_dir) / c.SOURCES_DIR / source.name / source.objectsLocation
    raw_files: List[Path] = [path for path in path.glob('**/*') if path.is_file()]
    target_features: List[m.FeatureProperties] = [f.properties for f in source.features]
    common_values: Dict[str, Any] = source.dict(include=m.RichManifest.construct().fields)

    with Pool(processes=cfg.parallelism or cpu_count()) as pool:
        tasks: Iterable[Union[m.Task, m.FeatureTask, m.UserFileTask]] = pool.imap_unordered(
            partial(_generate_task, user_files, features, path),
            raw_files
        )
        return m.TargetManifest(features=target_features, tasks=tasks, **common_values)


def cleanup_target_location(manifest_name: str, objects_dir: str, strict: bool):
    base_path: Path = Path(cfg.storage_dir) / c.PACKAGES_DIR / manifest_name

    if strict and base_path.exists():
        raise ManifestError(f"Base path {base_path} for modpack {manifest_name} already exists!")

    try:
        rmtree(base_path.as_posix())
    except FileNotFoundError as e:
        click.echo(click.style(f"Target {base_path.as_posix()} doesn't exist, suppressing…", fg="yellow", bold=True))
    except IOError as e:
        raise ManifestError(f"Can't clean target dir: {str(e)}") from e
    else:
        click.echo(f"Removed {click.style(base_path.as_posix(), bold=True)} (due to disabled strict mode)")

    try:
        base_path.mkdir()
        (base_path / objects_dir).mkdir()
    except IOError as e:
        raise ManifestError(f"Can't create target directory: {str(e)}") from e
    else:
        click.echo(f"Created empty directory: {click.style(base_path.as_posix(), bold=True)}")


def save_target_manifest(manifest: m.TargetManifest):
    base_path: Path = Path(cfg.storage_dir) / c.PACKAGES_DIR / manifest.name
    index_path: Path = base_path / c.TARGET_MANIFEST_INDEX

    try:
        json.dump(manifest.dict(), index_path.open("w", encoding="utf-8"), indent=2, ensure_ascii=False)
    except (IOError, ValueError, TypeError, json.decoder.JSONDecodeError) as e:
        raise ManifestError(f"Can't write target index: {str(e)}") from e
    else:
        click.echo(f"Wrote target index at: {click.style(index_path.as_posix(), bold=True)}")


def _save_target_task(
    source_base_path: Path,
    target_base_path: Path,
    debug: bool,
    task: Union[m.Task, m.UserFileTask, m.FeatureTask]
) -> bool:
    target: Path = target_base_path / task.location
    target.parent.mkdir(parents=True, exist_ok=True)
    try:
        copy(
            (source_base_path / task.to).as_posix(),
            (target_base_path / task.location).as_posix()
        )
    except IOError:
        return False
    else:
        return True
    finally:
        if debug:
            click.echo(click.style(f"Copying: {task.location} ← {task.to}", dim=True))


def save_target_content(source_manifest: m.SourceManifest, target_manifest: m.TargetManifest, debug: bool):
    source_path: Path = Path(cfg.storage_dir) / c.SOURCES_DIR / source_manifest.name / source_manifest.objectsLocation
    target_path: Path = Path(cfg.storage_dir) / c.PACKAGES_DIR / target_manifest.name / target_manifest.objectsLocation

    with Pool(processes=cfg.parallelism or cpu_count()) as pool:
        tasks: Iterable[bool] = pool.map(
            partial(_save_target_task, source_path, target_path, debug),
            target_manifest.tasks
        )
        result = all(tasks)

    if result:
        click.echo(click.style(f"Deployed {len(target_manifest.tasks)} files"))
    else:
        raise ManifestError("Error happened while copying target files")


def get_manifest_index_path() -> Path:
    return Path(cfg.storage_dir) / c.PACKAGES_DIR / cfg.modpack_index


def get_enabled_manifests() -> m.ManifestIndex:
    return m.ManifestIndex(**json.load(get_manifest_index_path().open()))


def save_manifests(modpack_index: m.ManifestIndex):
    with get_manifest_index_path().open('w') as file_handler:
        json.dump(modpack_index.dict(), file_handler, ensure_ascii=False, indent=2)


def get_manifest_info(
    source_manifest: Optional[m.SourceManifest],
    target_manifest: Optional[m.TargetManifest]
) -> m.ManifestInfo:
    if source_manifest is not None:
        location = Path(source_manifest.name) / c.TARGET_MANIFEST_INDEX
        name = source_manifest.name
        title = source_manifest.title
        version = source_manifest.version
    elif target_manifest is not None:
        location = Path(target_manifest.name) / c.TARGET_MANIFEST_INDEX
        name = target_manifest.name
        title = target_manifest.title
        version = target_manifest.version
    else:
        name = "unknown_" + uuid.uuid4().hex
        title = "unknown_" + uuid.uuid4().hex
        version = "0.0"
        location = Path("/tmp") / uuid.uuid4().hex

    return m.ManifestInfo(
        name=name,
        title=title,
        version=version,
        location=location.as_posix(),
    )


def bootstrap_source_manifest(manifest: m.SourceManifest) -> Path:
    base_path: Path = Path(cfg.storage_dir) / c.SOURCES_DIR / manifest.name

    if base_path.exists():
        raise ManifestError(f"Base path {base_path} for modpack {manifest.name} already exists!")

    base_path.mkdir()
    (base_path / manifest.objectsLocation).mkdir()

    index_path: Path = base_path / c.SOURCE_MANIFEST_INDEX

    yaml.dump(
        manifest.dict(),
        index_path.open("w", encoding="utf-8"),
        encoding="utf-8",
        indent=2,
    )

    return index_path



from itertools import chain
from pathlib import Path
from typing import Optional, List, Set

from apeiron import manifest
from apeiron.config import cfg
from apeiron import constants as c
from apeiron import exceptions as exc
from apeiron import models as m


def infer_modpack_status(source_manifest: m.SourceManifest, target_manifest: m.TargetManifest) -> m.ModpackStatus:
    if source_manifest is not None:
        if target_manifest is not None:
            if not manifest.validate_target_manifest(target_manifest):
                return m.ModpackStatus.BROKEN
            if manifest.build_target_manifest(source_manifest) == target_manifest:
                return m.ModpackStatus.FRESH
            else:
                return m.ModpackStatus.OUTDATED
        else:
            return m.ModpackStatus.NOT_DEPLOYED
    else:
        if target_manifest is not None:
            return m.ModpackStatus.ORPHANED
        else:
            return m.ModpackStatus.UNKNOWN


def get_modpack(package_id: str) -> m.Modpack:
    source_manifest: Optional[m.SourceManifest] = manifest.load_source_manifest(package_id)
    target_manifest: Optional[m.TargetManifest] = manifest.load_target_manifest(package_id)
    status: m.ModpackStatus = infer_modpack_status(source_manifest, target_manifest)
    manifest_info: m.ManifestInfo = manifest.get_manifest_info(source_manifest, target_manifest)

    return m.Modpack(
        name=package_id,
        status=status,
        enabled=manifest_info in manifest.get_enabled_manifests().packages,
        source_manifest=source_manifest,
        target_manifest=target_manifest,
    )


def get_modpacks() -> List[m.Modpack]:
    fs_modpack_ids: Set[str] = {
        d.name for d in
        chain(
            (Path(cfg.storage_dir) / Path(c.PACKAGES_DIR)).iterdir(),
            (Path(cfg.storage_dir) / Path(c.SOURCES_DIR)).iterdir(),
        )
        if d.is_dir()
    }

    enabled_modpacks_ids: Set[str] = {
        manifest_info.name
        for manifest_info
        in manifest.get_enabled_manifests().packages
    }

    return list(map(get_modpack, sorted(fs_modpack_ids | enabled_modpacks_ids)))


def enable_modpack(package_id: str):
    modpack: m.Modpack = get_modpack(package_id)

    if modpack.enabled:
        raise exc.StateConflictError(f"Modpack «{modpack.name}» already enabled!")

    if modpack.status not in {m.ModpackStatus.FRESH, m.ModpackStatus.OUTDATED}:
        raise exc.InvalidStatusError(
            f"Only «{m.ModpackStatus.FRESH.value}» and «{m.ModpackStatus.OUTDATED.value}» "
            f"can be enabled, but modpack «{modpack.name}» is «{modpack.status.value}»"
        )

    manifest_info: m.ManifestInfo = manifest.get_manifest_info(modpack.source_manifest, modpack.target_manifest)
    manifest_index: m.ManifestIndex = m.ManifestIndex(
        packages=[manifest_info, *manifest.get_enabled_manifests().packages]
    )

    manifest.save_manifests(manifest_index)


def disable_modpack(package_id: str):
    modpack: m.Modpack = get_modpack(package_id)

    if not modpack.enabled:
        raise exc.StateConflictError(f"Modpack «{modpack.name}» already disabled!")

    manifest_index: m.ManifestIndex = m.ManifestIndex(packages=[
        manifest_info
        for manifest_info
        in manifest.get_enabled_manifests().packages
        if manifest_info.name != package_id
    ])

    manifest.save_manifests(manifest_index)


def delete_modpack(package_id: str, prune: bool = False) -> List[exc.ManifestError]:
    modpack: m.Modpack = get_modpack(package_id)
    warnings: List[exc.ManifestError] = []

    if modpack.enabled:
        raise exc.StateConflictError(f"Can't delete modpack «{modpack.name}», because it's enabled!")

    if modpack.target_manifest is None:
        warnings.append(exc.ConsistenceError(f"There are no built files for «{modpack.name}», suppressing…"))
    else:
        manifest.delete_target_manifest(modpack.target_manifest)

    if prune:
        if modpack.source_manifest is None:
            warnings.append(exc.ConsistenceError(f"There are no sources for «{modpack.name}», suppressing…"))
        else:
            manifest.delete_source_manifest(modpack.source_manifest)

    return warnings


def build_modpack(package_id: str, strict: bool, debug: bool):
    modpack: m.Modpack = get_modpack(package_id)
    allowed_statuses: Set[m.ModpackStatus] = {
        m.ModpackStatus.FRESH,
        m.ModpackStatus.OUTDATED,
        m.ModpackStatus.NOT_DEPLOYED,
        m.ModpackStatus.BROKEN,
    }

    if modpack.status not in allowed_statuses:
        statuses = ', '.join(s.value for s in allowed_statuses)
        raise exc.InvalidStatusError(
            f"Only {statuses} allowed to build! "
            f"Modpack «{modpack.name}» is {modpack.status.value} and should be only deleted."
        )

    built_manifest: m.TargetManifest = manifest.build_target_manifest(modpack.source_manifest)

    manifest.cleanup_target_location(built_manifest.name, built_manifest.objectsLocation, strict)
    manifest.save_target_manifest(built_manifest)
    manifest.save_target_content(modpack.source_manifest, built_manifest, debug)
